#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <alloca.h>


#define MatchCost 5
#define MismatchCost -3
#define InsertionCost -3
#define DeletionCost -5

int main(int argc, char * argv[])
{
	const char * restrict Sequence;
	static const char Alphabet[] = "ACGT";
	
	if (argc < 2) {
		fprintf(stderr, "Usage : %s sequence [at_start]\n", argv[0]);
		return 1;
	}
	
	Sequence = argv[1];
	const size_t SequenceLength = strlen(Sequence);
	
	/***************************************************************************************
	 * Check that sequence only has DNA
	 ***************************************************************************************/
	for (size_t i=0; i<SequenceLength; i++) {
		switch(Sequence[i]) {
			case 'A':
			case 'C':
			case 'G':
			case 'T':
				break;
			default:
				fprintf(stderr,"Your sequence contains non DNA symbols\n");
				return 1;
		}
	}
	
	/***************************************************************************************
	 * Generate header
	 ***************************************************************************************/
	printf("ID   GenPRF_%32s; MATRIX.\n"
	       "MA   /GENERAL_SPEC: ALPHABET='%s'; LENGTH=%zu;\n"
				 "MA   /DISJOINT: DEFINITION=PROTECT; N1=2; N2=%zu\n",
				Sequence, Alphabet, SequenceLength, SequenceLength-1);
	
	/* Normalization and cutoffs */
	printf("MA   /NORMALIZATION: MODE=1; FUNCTION=LINEAR; R1=0.0; R2=0.1; TEXT='No_units';\n"
				 "MA   /CUT_OFF: LEVEL=-1; SCORE=65; N_SCORE=6.5; MODE=1; TEXT='?';\n"
				 "MA   /CUT_OFF: LEVEL=0; SCORE=85; N_SCORE=8.5; MODE=1; TEXT='!';\n");

	/***************************************************************************************
	 * Matrix scores
	 ***************************************************************************************/
	/* Default values */
	if (argc != 3) {
		printf("MA   /DEFAULT: BM=*; ME=*; DM=0; IM=0; MD=-1; MI=-1; M0=-1; D=%i; I=%i; II=-1; DD=-2;\n",
					 DeletionCost, InsertionCost);
	}
	else {
		printf("MA   /DEFAULT: B1=*; BM=*; ME=*; DM=0; IM=0; MD=-1; MI=-1; M0=-1; D=%i; I=%i; II=-1; DD=-2;\n",
				 DeletionCost, InsertionCost);
	}
	/* First line */
	if (argc != 3) {
		printf("MA   /I: BM=0; BD=0; E1=*;\n");
	}
	else {
		printf("MA   /I: BM=0; BI=0; BD=0; E1=*;\n");
	}
	
	/* Middle lines */
	for (size_t i=0; i<SequenceLength; i++) {
		
		/* Match score */
		printf("MA   /M: SY='%c'; M=", (int) Sequence[i]);
		for (size_t j=0; j<3; j++) {
				printf("%i,", (Sequence[i] == Alphabet[j]) ? MatchCost: MismatchCost);
		}
		printf("%i;", (Sequence[i] == Alphabet[3]) ? MatchCost: MismatchCost);
		
		/* Deletion score */
		if (i<SequenceLength-1) {
			printf(" D=%i;\n", (Sequence[i] == Sequence[i+1]) ? 0 : DeletionCost);
		}
		else {
			printf(" D=%i;\n", DeletionCost);
		}
		
		/* Insertion scores */
		if (i < SequenceLength-1) {
			printf("MA   /I: I=");
			for (size_t j=0; j<3; j++) {
					printf("%i,", (Sequence[i] == Alphabet[j]) ? -1: InsertionCost);
			}
			printf("%i;\n", (Sequence[i] == Alphabet[3]) ? -1: InsertionCost);
		}
	}
	
	/* Last line */
	printf("MA   /I: B1=*; ME=0; DE=0; \n");
	
	/***************************************************************************************
	 * Generate footer
	 ***************************************************************************************/
	printf("CC   Generated by GenPrf\n//\n");
	
	return 0;
}
